/*
Copyright (c) 2018 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdint.h>
#include <ft2build.h>
#include FT_FREETYPE_H
#include <math.h>
#include "font.h"
#include "womap.h"

static  FT_Library ftlib;
static int font_dpi;

static char *metrics_chars= "MWm{|Q^z#$`gj@";

typedef struct
{
  int  init;
  FT_Face face;
  womap_t *glyphs;
  int width, height, ascent;
} font_t;

static font_t fonts[NFONTS];

static  char *fterrstr(int code) ;

static glyph_t bad_prg, bad_chr, fon_err;

static int imin2(int a,int b) { return a<b?a:b; }
static int imax2(int a,int b) { return a>b?a:b; }

static void pat2glyph(char *data, glyph_t *G,int w,int h,int a)
{
  int r,c;
  G->width= w;
  G->height= h;
  G->ascent= a;
  if (G->data) free(G->data);
  G->data= calloc(1,w*h);
  for(r=0;r<h && r<data[1];r++)
    for(c=0;c<w && c<data[0];c++)
      G->data[r*w+c]= (data[r*data[0]+c+2] ? 255 : 0);
}

static char *BAD[]= 
{
"\020\020\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\1\1\1\0\0\0\0\1\0\0\0\0\0\1\0\0\1\0\0\1\0\0\1\0\1\0\0\0\0\1\0\0\1\0\0\1\0\1\0\0\0\1\0\0\0\1\0\0\1\1\1\0\0\1\0\0\0\1\0\0\1\1\0\0\1\0\0\1\0\1\1\1\1\1\0\1\0\1\0\0\1\0\0\1\0\1\0\0\0\1\0\1\0\1\0\0\1\1\1\0\0\1\0\0\0\1\0\0\1\1\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\1\1\1\0\0\1\1\1\0\0\0\1\1\1\0\0\1\0\0\1\0\1\0\0\1\0\1\0\0\0\0\0\1\0\0\1\0\1\0\0\1\0\1\0\0\0\0\0\1\1\1\0\0\1\1\1\0\0\1\0\1\1\0\0\1\0\0\0\0\1\0\0\1\0\1\0\0\1\0\0\1\0\0\0\0\1\0\0\1\0\0\1\1\1\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0",
"\020\020\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\1\1\1\0\0\0\1\1\0\0\1\1\0\0\0\0\1\0\0\1\0\1\0\0\1\0\1\0\1\0\0\0\1\0\0\1\0\1\0\0\1\0\1\0\0\1\0\0\1\1\1\0\0\1\1\1\1\0\1\0\0\1\0\0\1\0\0\1\0\1\0\0\1\0\1\0\0\1\0\0\1\0\0\1\0\1\0\0\1\0\1\0\0\1\0\0\1\1\1\0\0\1\0\0\1\0\1\1\1\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\1\1\0\0\1\0\0\1\0\1\1\1\0\0\0\1\0\0\1\0\1\0\0\1\0\1\0\0\1\0\0\1\0\0\0\0\1\0\0\1\0\1\0\0\1\0\0\1\0\0\0\0\1\1\1\1\0\1\1\1\0\0\0\1\0\0\0\0\1\0\0\1\0\1\0\0\1\0\0\1\0\0\1\0\1\0\0\1\0\1\0\0\1\0\0\0\1\1\0\0\1\0\0\1\0\1\0\0\1\0",
"\020\020\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\1\1\1\1\0\0\1\1\0\0\1\0\0\0\1\0\1\0\0\0\0\1\0\0\1\0\1\1\0\0\1\0\1\0\0\0\0\1\0\0\1\0\1\0\1\0\1\0\1\1\1\0\0\1\0\0\1\0\1\0\1\0\1\0\1\0\0\0\0\1\0\0\1\0\1\0\1\0\1\0\1\0\0\0\0\1\0\0\1\0\1\0\0\1\1\0\1\0\0\0\0\0\1\1\0\0\1\0\0\0\1\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\1\1\1\1\0\0\0\0\0\0\0\0\0\0\0\0\1\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\1\0\0\0\0\0\1\1\0\0\1\1\0\0\0\0\1\1\1\0\0\1\0\0\0\1\0\0\0\0\0\0\1\0\0\0\0\1\0\0\0\1\0\0\0\0\0\0\1\0\0\0\0\1\0\0\0\1\0\0\0\0\0\0\1\1\1\1\0\1\0\0\0\1\0\0\0\0\0",
};
int font_init(int dpi)
{
  int i;
  for(i=0;i<NFONTS;i++) { fonts[i].glyphs= womap_new(); fonts[i].init= 0; }
  font_dpi= dpi;
  pat2glyph( BAD[0], &bad_prg, 16,16,16);
  pat2glyph( BAD[1], &bad_chr, 16,16,16);
  pat2glyph( BAD[2], &fon_err, 16,16,16);
  return 0;
}


static int get_char_metrics
   (FT_Face face, unsigned int ch,int *W,int *A,int *D)
{
   unsigned int gindex;
   int width,height,ascent;
   FT_GlyphSlot slot;
   FT_Bitmap *ftbitmap; 
   int left_padding;
   int rwidth;        /* R stands for rendered */

   gindex= FT_Get_Char_Index(face, ch);
   if (!gindex) return 1;
   if (FT_Load_Glyph(face, gindex, FT_LOAD_RENDER)) return 1;
   slot= face->glyph;
   ftbitmap= &slot->bitmap;

   width= slot->metrics.horiAdvance/64;
   left_padding= slot->bitmap_left;
   ascent= slot->bitmap_top;
   height= ftbitmap->rows;
   rwidth= ftbitmap->width;

   /* sometimes chars are designed to overflow the given box */
   if (left_padding<0) left_padding=0;
   /* same as above, from right */
   if (width<(rwidth+left_padding)) width= rwidth+left_padding;

   *W= width;
   *A= ascent;
   *D= height-ascent;

   return 0;
}

glyph_t *font_render(int fno, unsigned int ch)
{
   font_t *FF;
   unsigned int gindex;
   int error;
   FT_GlyphSlot slot;
   FT_Bitmap *ftbitmap; 
   int i,j,k;
   glyph_t *G;

   unsigned char *ddata;
   unsigned char *sdata;
   int dw, dh, dpitch,      // dest width, height, pitch
       sw, sh, spitch;      // src  width, height, pitch
   int Cw, Ch;              // copy width, height

   FF= fonts+fno;
   if (!FF->init) return &bad_prg;

   G= womap_find(FF->glyphs, ch);
   if (G) return G;

   gindex= FT_Get_Char_Index(FF->face, ch);
   if (!gindex) 
   {
      //fprintf(stderr,"character U%x\n doesn't exist in font.\n", ch);
      return &bad_chr;
   }

   error= FT_Load_Glyph(FF->face, gindex, FT_LOAD_RENDER);
   if (error)
   {
      //fprintf(stderr, "could not load or render character U%x\n"
      //               "freetype error: %s\n", (unsigned)ch, fterrstr(error));
      return &fon_err;
   }

   slot= FF->face->glyph;
   ftbitmap= &slot->bitmap;

   G= malloc(sizeof(*G));
   G->width= FF->width;
   G->height= FF->height;
   G->ascent= slot->bitmap_top;
   G->data= calloc(1,FF->width*FF->height);

   dw= FF->width;       dh= FF->height;     dpitch= FF->width;
   sw= ftbitmap->width; sh= ftbitmap->rows; spitch= ftbitmap->pitch;

   ddata= G->data;      sdata= ftbitmap->buffer;

   k= FF->ascent-slot->bitmap_top;
   if (k<0)      { k= -k; sh-= k; sdata+= k*spitch; }
   else if (k>0) {        dh-= k; ddata+= k*dpitch; }

   /* sometimes chars are designed to overflow the given box, that's
      the reason for  the imax2 call.  */
   if (dw>sw) ddata+= imin2(imax2(slot->bitmap_left,0), dw-sw);

   Cw= imin2(sw,dw);
   Ch= imin2(sh,dh);

   for(i=0;i<Ch;i++) 
      for(j=0;j<Cw;j++)
         ddata[j+dpitch*i] = sdata[j+spitch*i];

   womap_insert(FF->glyphs, ch, G);
   return G;
}

static int initlib()
{
   int error;
   static int lib_init= 0;
   if (!lib_init)
   {
      lib_init= 1;
      error= FT_Init_FreeType( &ftlib );
      if (error) 
      {
         fprintf(stderr, "could not initialize freetype library\n"
                         "freetype error: %s\n", fterrstr(error));
         return 1;
      }
   }
  return 0;
}

static void set_font(int fno,FT_Face face)
{
  font_t *f;
  glyph_t *g;
  int i, maxA, maxD, maxW, A,D,W;
  f= fonts+fno;
  if (f->init)
    FT_Done_Face(f->face);
  f->face= face;
  while((g=womap_destroy(f->glyphs))) 
  {
    free(g->data); free(g);
  }
  maxA= maxD= maxW= 0;
  for(i=0;metrics_chars[i];i++)
  {
    if (get_char_metrics(face, metrics_chars[i], &W, &A, &D)) continue;
    maxA= imax2(maxA, A);
    maxD= imax2(maxD, D);
    maxW= imax2(maxW, W);
  }
  if (!maxW) maxW= 16;
  if (!maxA) maxA= 16;
  if (!maxD) maxD= 5;
  f->width= maxW;
  f->height= maxA+maxD;
  f->ascent= maxA;
  f->init= 1;
}

int font_load_mem(int fno, const unsigned char *data, size_t dsize, int fsize)
{
   int error;
   FT_Face face;
   if (initlib()) return 1;
   error= FT_New_Memory_Face(ftlib, data, dsize, 0, &face);
   if (error)
   {
       fprintf(stderr, "can not load face #0 from memory %p\n"
	               "freetype error: %s\n", data, fterrstr(error));
       return 1;
   }
	/* FIXME: handle the error properly */
   FT_Set_Char_Size( face, 0, fsize * 64, font_dpi-font_dpi/8, font_dpi);
   set_font(fno, face);
   return 0; 
}

int font_load_file(int fno,char *file_name,int size)
{
   int error;
   FT_Face face;
   if (initlib()) return 1;
   error= FT_New_Face(ftlib, file_name, 0, &face);
   if (error)
   {
       fprintf(stderr, "can not load face #0 from file %s\n"
	               "freetype error: %s\n", file_name, fterrstr(error));
       return 1;
   }
	/* FIXME: handle the error properly */
   FT_Set_Char_Size( face, 0, size * 64, 0, font_dpi);
   set_font(fno, face);
   return 0; 
}

static  char * fterrstr(int code)
{
   static char buf[200];
#define FT_ERROR_START_LIST     switch(code) {
#define FT_ERRORDEF( e, v, s )  case e: return s;
#define FT_ERROR_END_LIST       }
#include FT_ERRORS_H
   sprintf(buf," Unknown error %d",code);
   return buf;
}

int font_dimensions(int fno, int *W, int *H, int *A)
{
  font_t *FF;
  FF= fonts+fno;
  if (!FF->init) return 1;
  if (W) *W= FF->width;
  if (H) *H= FF->height;
  if (A) *A= FF->ascent;
  return 0;
}

static void make_light_boxdraw(font_t *F);

int font_set_cell(int fno, int W, int H, int A)
{
  font_t *FF;
  FF= fonts+fno;
  if (!FF->init) return 1;
  FF->width= W;
  FF->height= H;
  FF->ascent= A;
  pat2glyph( BAD[0], &bad_prg, W,H,A);
  pat2glyph( BAD[1], &bad_chr, W,H,A);
  pat2glyph( BAD[2], &fon_err, W,H,A);
  make_light_boxdraw(FF);
  return 0;
}




/**
  Generally speaking, box drawing characters are often problematic with
  truetype fonts. In the best case, everything gets aligned, but there
  are gaps at the edges of horizontal/vertical lines, which is OK but
  doesn't look as good. Therefore, I'm going to patch the font at runtime,
  using the final values for the font width and height.

  Now, there are a bunch of these. I will make the light ones first, and
  then see if there is a need to do the heavier/dashed stuff.

  Light:
    2500        horizontal                 2502        vertical
    250C        upper left corner          2510        upper right corner
    2514        lower left corner          2518        lower right corner
    251C        tee pointing right         2524        tee pointing left
    252C        tee pointing down          2534        tee pointing up
    253C        cross

 These can be rendered as a combination of the following line segments:

                2
                |
            1 =   =  4
                |
                8

 So, here is the table.
**/
static int light_boxdraw[]= {
  0x2500,  5, 0x2502, 10, 0x250C, 12, 0x2510, 9, 0x2514, 6, 0x2518, 3,
  0x251C, 14, 0x2524, 11, 0x252C, 13, 0x2534, 7, 0x253C, 15, 0, 0 }; 

static void light_horzline(glyph_t *G,int x1,int x2,int y)
{
  int x;
  for(x=x1;x<=x2;x++) G->data[y*G->width+x]= 0xff;
}

static void light_vertline(glyph_t *G, int x, int y1, int y2)
{
  int y;
  for(y=y1;y<=y2;y++) G->data[y*G->width+x]= 0xff;
}

static glyph_t* mkaltgl(font_t *F, int code)
{
  glyph_t *G;
  G= malloc(sizeof(*G));
  G->width= F->width; G->height= F->height; G->ascent= F->ascent;
  G->data= calloc(1,G->width*G->height);
  womap_insert(F->glyphs, code, G);
  return G;
}

static void light_linedraw(glyph_t *G, int segs)
{
  int Mx, My;
  Mx= G->width / 2;
  My= G->height / 2;
  if (segs&1) light_horzline(G, 0,          Mx,          My);
  if (segs&4) light_horzline(G, Mx, G->width-1,          My);
  if (segs&2) light_vertline(G, Mx,          0,          My);
  if (segs&8) light_vertline(G, Mx,         My, G->height-1);
}

static void make_light_boxdraw(font_t *F)
{
  int i;
  for(i=0;light_boxdraw[i];i+=2)
    light_linedraw(mkaltgl(F, light_boxdraw[i]), light_boxdraw[i+1]);
}
