#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include "matrix.h"
#include "x11.h"

static char *help_lines[]=
{
"  VERSION                                                                   ",
"      muterm-1.7  02 June 2021                                              ",
"                                                                            ",
"  TABS                                                                      ",
"     =1=   current   2   background  [3] inactive                           ",
"     Type control-digit to switch to the given tab.                         ",
"     control-right and control-left travel between tabs.                    ",
"                                                                            ",
"  SELECTIONS                                                                ",
"     Selecting text puts it in the PRIMARY selection.                       ",
"     control-space moves it into the CLIPBOARD selection and frees up       ",
"     the PRIMARY selection.                                                 ",
"     Middle mouse button or shift-insert pastes PRIMARY selection.          ",
"     control-insert pastes CLIPBOARD selection.                             ",
"     If program hangs due to the selection owner not responding, use        ",
"     the window manager's close button (x) to cancel the operation.         ",
"                                                                            ",
"  SCROLL BUFFER                                                             ",
"     shift-pageup, shift-pagedown, mouse scroll wheel is used to view       ",
"     the scroll buffer. Typing a key will bring the normal view back.       ",
"                                                                            ",
"  SHORTCUTS                                                                 ",
"     In fullscreen mode, mouse scroll wheel is a shortcut for pageup/       ",
"     pagedown keys.                                                         ",
"                                                                            ",
"  FONT                                                                      ",
"     The font used is based on M+ fonts. Here is the text attached to it    ",
"     in the distribution.                                                   ",
"                                                                            ",
"----------------------------------------------------------------------------",
"M+ FONTS                            Copyright (C) 2002-2017 M+ FONTS PROJECT",
"                                                                            ",
" LICENSE_E                                                                  ",
"                                                                            ",
"These fonts are free software.                                              ",
"Unlimited permission is granted to use, copy, and distribute them, with     ",
"or without modification, either commercially or noncommercially.            ",
"THESE FONTS ARE PROVIDED \"AS IS\" WITHOUT WARRANTY.                          ",
"                                                                            ",
"http://mplus-fonts.osdn.jp                                                  ",
"----------------------------------------------------------------------------",
"                                                                            ",
"  DISTRIBUTION                                                              ",
"                                                                            ",
"Copyright (c) 2018-2021 Sukru Cinar.                                        ",
"                                                                            ",
"All rights reserved.                                                        ",
"                                                                            ",
"Redistribution and use in source and binary forms, with or without          ",
"modification, are permitted provided that the following conditions          ",
"are met:                                                                    ",
"                                                                            ",
"1. Redistributions of source code must retain the above copyright notice,   ",
"   this list of conditions and the following disclaimer.                    ",
"                                                                            ",
"2. Redistributions in binary form must reproduce the above copyright        ",
"   notice, this list of conditions and the following disclaimer in the      ",
"   documentation and/or other materials provided with the distribution.     ",
"                                                                            ",
"THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS         ",
"\"AS IS\" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT         ",
"LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR       ",
"A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT        ",
"HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,      ",
"SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED    ",
"TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR      ",
"PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF      ",
"LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING        ",
"NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS          ",
"SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                ",
"                                                                            ",
NULL
};

static int hestart= 0;

static void frame_title(int row,char *str)
{
  int sta;
  int i;
  sta= 77-strlen(str);
  
  for(i=0;str[i];i++)
  {
    display_glyph(row, sta+i, CELLATTR_BOLD | str[i]);
  }
}

static void print_border()
{
  int i;
  display_glyph( 0,  0, 0x250C);
  display_glyph( 0, 79, 0x2510);
  display_glyph(23,  0, 0x2514);
  display_glyph(23, 79, 0x2518);
  for(i=1;i<79;i++)
  {
    display_glyph( 0, i, 0x2500);
    display_glyph(23, i, 0x2500);
  }
  for(i=1;i<23;i++)
  {
    display_glyph(i,  0, 0x2502);
    display_glyph(i, 79, 0x2502);
  }
  frame_title(0, "MUTERM HELP");
  frame_title(23, "q/up/down");
}

static void print_help_line(int row,int col, char *str)
{
  int c;
  for(c=0;str[c];c++)
  {
    display_glyph(row, c+col, (uint32_t) str[c]);
  }
  for(;c<78;c++) display_glyph(row, c+col, ' ');
}

static void help_print()
{
  int i;
  for(i=0;i<22;i++)
    print_help_line(i+1,1,help_lines[hestart+i]);
  update_area(0,0,80,24);
}


static int stop_help;


void help_show()
{
  stop_help= 0;
  set_cursor(-1,-1);
  print_border();
  hestart= 0;
  help_print();   
  while(!stop_help)
  {
    gfx_flush();
    gfx_event();
  }
}


void help_key(int key, unsigned int wide, char *str)
{
  int end,i,hh;
  switch(key)
  {
  case GFX_K_PRT:
    if (wide=='q') stop_help= 1;
    break;
  case GFX_K_UP:
    hh= hestart;
    hestart-= 21;
    if (hestart<0) hestart= 0;
    if (hestart!=hh) help_print();
    break;
  case GFX_K_DOWN:
    end= hestart+21;
    for(i=0;i<21;i++)
      if (!help_lines[end+i]) break;
    hestart= end+i-22;
    help_print();
    break;
  }
}
