/*
Copyright (c) 2018 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

// #define CELLATTR_SEL   (1<<21)

#define CELLATTR_ULINE (1<<22)
#define CELLATTR_BLINK (1<<23)
#define CELLATTR_REV   (1<<24)
#define CELLATTR_BOLD  (1<<25)
#define CELLATTR_HIDDEN (1<<27)

// unicode characters run up to E_00_7f, so 20 bits is enough
#define CELL_CHRMASK   ((1<<21)-1)

/* Each cell is represented by a 32 bit integer. The lower 20 bits 
   identify the glyph. The higher ones are set by the flags above. */


void display_glyph(int row, int col, uint32_t C);
void set_cursor(int row, int col);
void set_sel(int br, int bc, int er, int ec);
void update_area(int col, int row, int wid, int hei);
void set_title(uint8_t *S);
void have_selection(uint32_t *s);

typedef struct screen screen_t;

screen_t *screen_new(int nrows,int ncols);
void screen_print_str(screen_t *S,const uint8_t *str, int len);
void screen_updates(screen_t *S);
void screen_redraw(screen_t *S);

void screen_page_prev(screen_t *S);
void screen_page_next(screen_t *S);
void screen_page_last(screen_t *S);

void screen_prelease(screen_t *S);
void screen_pmove(screen_t *S,int row, int col);
void screen_ppress(screen_t *S,int row,int col);
void screen_select_none(screen_t *S);

void screen_free(screen_t *S);

int screen_bs_mode(screen_t *S);
int screen_del_mode(screen_t *S);

void screen_set_fd(screen_t *S, int fd);

void screen_activate(screen_t *S,int value);

void help_show();
void help_key(int key, unsigned int wide, char *str);

void matrix_debug(int d);
void display_modified(); // time to call gfx_flush
int screen_mode(screen_t *S);
