/*
Copyright (c) 2018 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

typedef struct 
{
  /** when outlen is non-zero, you may use outbuf[] to get
      the output. after this, don't forget to call _reset()
      so that the decoder can be used one more time **/
  unsigned int outbuf[10];
  int outlen;

  /*** private fields follow ***/
  unsigned char pbuf[10];
  int size;
  int len;
  unsigned int erract, accu;
} utf8decoder_t;

#define UTF8_DELETE  0  // deletes any invalid chars from the stream 
#define UTF8_REPORT  1  // keeps invalid char sequences but marks them 
                        // as invalid by setting bit31
/*****
   erract is one of UTF8_DELETE, UTF8_REPORT or a replacement integer value.
   if a replacement value is given, all invalid characters will be replaced
   by that value when an error occurs.
 *****/
void utf8decoder_init(utf8decoder_t *dec,int erract);

/*****
   reset to the initial state.
 *****/
void utf8decoder_reset(utf8decoder_t *dec);

/*****
   feed one character to the machine. the return value
   is the number of integers in the output buffer. for
   normal operation, this is just 1. when there is an
   error, we can have up to 6 erronous words in the output.
   returns zero if the sequence isn't complete yet.
 *****/
int  utf8decoder_chr(utf8decoder_t *dec,unsigned char chr);

/*****
   signal the end of stream. return value has the same
   semantics as _chr(), but this function can only return
   invalid bytes since a valid sequence would have been
   already reported by _chr().
 *****/
int  utf8decoder_end(utf8decoder_t *dec);

