/*
Copyright (c) 2018 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdlib.h>
#include <errno.h>
#include <sys/select.h>
#include <stdint.h>
#include "unisel.h"


typedef struct { int how; int fd; int code; } selent_t;

#define EV(x,y)  (((x)<<8)|(y))

struct selector
{
  int nrec, srec;
  selent_t *rec;
  uint16_t *evt;
  int sevt, nevt, cur;
};



SEL *SELnew()
{
  SEL *S;
  S= calloc(1,sizeof(*S));
  S->srec= 10;
  S->rec= malloc(sizeof(S->rec[0])*S->srec);
  return S;
}

void SELfre(SEL* S)
{
  free(S->rec);
  free(S);
}


int  SELadd(SEL* S,int fd, int code, int how)
{
  int i;
  if (fd<0) return fd;
  how &= SELa;
  if (!how) return -1;
  for(i=0;i<S->nrec;i++)
   if (S->rec[i].fd==fd || S->rec[i].code==code)
     break;
  if (i==S->nrec)
  {
    if (S->nrec==S->srec)
    {
      S->srec+= S->srec/2 + 5;
      S->rec= realloc(S->rec, sizeof(S->rec[0])*S->srec);
    }
    S->nrec++;
  } 
  S->rec[i].fd= fd;
  S->rec[i].code= code;
  S->rec[i].how= how;
  return fd;
}

int  SELrem(SEL* S,int fd, int code)
{
  int i;
  int Rfd;
  if (!S->nrec) return fd;
  Rfd= fd;
  for(i=0;i<S->nrec;i++)
   if ((fd==-1 || S->rec[i].fd==fd) &&
       (code==-1 || S->rec[i].code==code))
   {
     Rfd= S->rec[i].fd;
     S->rec[i]= S->rec[S->nrec-1];
     S->nrec--;
     i--;
   }
  return Rfd;
}

static void put_fds(SEL *S, fd_set *set,int how)
{
  int i,fd;
  for(i=0;i<S->nrec;i++)
   if (S->rec[i].how & how)
   {
     fd= S->rec[i].fd;
     if (!FD_ISSET(fd, set)) continue;
     S->evt[S->nevt++]= EV(how,S->rec[i].code);
   }
}

int  SELbeg(SEL* S,int mstimeout)
{
  int i, ho, fd, R;
  int nr, nw, nx;
  struct timeval *tv, tvstore;
  fd_set rset, wset, xset;
  int maxfd;

  if (S->sevt<3*S->nrec)
  {
    S->sevt= 3*(S->nrec+3);
    S->evt= realloc(S->evt, sizeof(S->evt[0])*S->sevt);
  }
  S->nevt= 0;

again:
  FD_ZERO(&rset);
  FD_ZERO(&wset);
  FD_ZERO(&xset);

  nr= nw= nx= 0;
  maxfd= -1;
  for(i=0;i<S->nrec;i++)
  {
    ho= S->rec[i].how;
    fd= S->rec[i].fd;
    if (ho&SELr) { FD_SET(fd, &rset); nr++; }
    if (ho&SELw) { FD_SET(fd, &wset); nw++; }
    if (ho&SELx) { FD_SET(fd, &xset); nx++; }
    if (fd>maxfd) maxfd= fd; 
  }


  if (mstimeout<0) // no timeout
  {
    tv= NULL;
  }
  else
  {
    tv= &tvstore;
    tvstore.tv_sec= mstimeout / 1000;
    tvstore.tv_usec= (mstimeout % 1000) * 1000;
  }
  R= select(maxfd+1, 
              nr ? &rset : NULL, 
              nw ? &wset : NULL, 
              nx ? &xset : NULL, tv);

  if (R<0 && errno==EINTR) goto again;
  if (R<=0)
  {
     return SELend;
  }  
  if (nx)
  {
    for(i=0;i<S->nrec;i++)
    {
      fd= S->rec[i].fd;
      if (FD_ISSET(fd, &xset))
      {
        FD_CLR(fd, &rset);
        FD_CLR(fd, &wset);
      }
    }
  }
  put_fds(S, &xset, SELx);
  put_fds(S, &rset, SELr);
  put_fds(S, &wset, SELw);
  if (S->nevt==0) return SELend;
  S->cur= 0;
  return S->evt[S->cur];
}

int  SELnex(SEL* S)
{
  if (S->cur>=S->nevt-1) return SELend;
  S->cur++;
  return S->evt[S->cur];
}
